

fun printVerbose ss = if !VERBOSE then ErrPrint.print ss else ();

(* returns value associated with x in l, or x itself if none *)
(* fun alias x l = case List.find (fn (z,_) => z=x) l of SOME (_,y) => y | _ => x; *)
fun alias x ((y,z)::t) = if x=y then z else alias x t | alias x [] = x;


(* error messages : + error defined in frac-flags *)

fun notyet m = error ("feature not supported yet (" ^ m ^ "), sorry");
fun comperror m = error ("cannot compile: " ^ m);
fun compwarning m r = (warning ("warning: " ^ m); r);
fun failwith m = error ("unexpected failure (" ^ m ^ "), please report it");


(* for emit *)

fun pow (x,0) = 1.0 | pow (x,1) = x | pow (x,y) = pow (10.0 * x,y-1);

(* for totts, etc *)

fun range(min,max) = if min=max then [max] else min :: (range(min+1,max))

(* applies f to n-1 .. 0 *)
(* fun iter 0 f = () | iter 1 f = f 0 | iter n f = (f (n-1); iter (n-1) f); *)
fun iter n f = let fun g n = if n>0 then (f (n-1); g (n-1)) else () in g n end;

(* index of x in list l (from 0) [fails with Match if not member] *)
fun index x l =
    let fun rind (h::t) n = if h=x then n else rind t (n+1)
    in rind l 0
    end;

(* union of ordered sets, sup is comparison funtion: *)
fun ounion sup s1 s2 =
    let fun insert x [] = [x]
    	  | insert x (l as (h::t)) = if x=h then l else if sup (x,h) then h::insert x t else x::l
	fun union [] s2 = s2 | union (h::t) s2 = insert h (union t s2)
    in union s1 s2
    end;

 
(* BitArray/BitVector structures in terms of CharARray/CharVector *)

signature BITVECTOR = sig
type vector
val tabulate : int * (int -> bool) -> vector
val sub : vector * int -> bool
val length : vector -> int
val fromList : bool list -> vector
val appi : (int * bool -> unit) -> vector -> unit
val foldr : (bool * 'a -> 'a) -> 'a -> vector -> 'a
end;

structure BitVector : BITVECTOR = struct
type vector = CharVector.vector;
fun tabulate (n,f) = CharVector.tabulate(n,fn i => Char.chr (if f i then 1 else 0));
fun sub (a,i) = CharVector.sub (a,i) = Char.chr 1
fun length a = CharVector.length a;
fun fromList l = CharVector.fromList (List.map (fn b => Char.chr (if b then 1 else 0)) l);
fun appi f v = CharVector.appi (fn (i,c) => f (i,c = Char.chr 1)) v;
fun foldr f a v = CharVector.foldr (fn (c,g) => f (c = Char.chr 1,g)) a v;
end;

signature BITARRAY = sig
type array
val tabulate : int * (int -> bool) -> array
val sub : array * int -> bool
val update : array * int * bool -> unit
val length : array -> int
val array : int * bool -> array
val fromList : bool list -> array
val appi : (int * bool -> unit) -> array -> unit
val foldr : (bool * 'a -> 'a) -> 'a -> array -> 'a
val vector : array -> BitVector.vector
end;

structure BitArray : BITARRAY = struct
type array = CharArray.array;
fun tabulate (n,f) = CharArray.tabulate(n,fn i => Char.chr (if f i then 1 else 0));
fun sub (a,i) = CharArray.sub (a,i) = Char.chr 1
fun update (a,i,b) = CharArray.update(a,i,Char.chr(if b then 1 else 0));
fun length a = CharArray.length a;
fun array (n,b) = CharArray.array(n,Char.chr(if b then 1 else 0));
fun fromList l = CharArray.fromList (List.map (fn b => Char.chr (if b then 1 else 0)) l);
fun appi f v = CharArray.appi (fn (i,c) => f (i,c = Char.chr 1)) v;
fun foldr f a v = CharArray.foldr (fn (c,g) => f (c = Char.chr 1,g)) a v;
fun vector a = CharArray.vector a;
end;

