

(* prints the source by Prn *)
functor SrcPrint (Prn : PRN) = struct

open Prn;

fun newline n = (print "\n"; repeat n print " ");

fun print_Fiacre (Program dl) =
        List.app (fn TypeDecl d => print_TypeDecl 0 d
	           | ChannelDecl d => print_ChannelDecl 0 d
	           | ConstantDecl d => print_ConstantDecl 0 d
	           | FunctionDecl d => print_FunctionDecl 0 d
	           | ExternDecl d => print_ExternDecl 0 d
	           | ProcessDecl d => print_ProcessDecl 0 d
	           | ComponentDecl d => print_ComponentDecl 0 d
		   | Main (_,id,lower) => (newline 0; print id; case lower of [] => () | _ => (print " > "; print_list (""," | ","") print lower; newline 0))
		   | PropertyDecl d => print_PropertyDecl 0 d
		   | Assert (_,id) => (newline 0; print "assert "; print id)) dl
and print_TypeDecl c (l,a,b) =
	(newline c; print "type "; print a; print " is "; print_Type b; newline 0)
and print_ChannelDecl c (l,a,b) =
	(newline c; print "channel "; print a; print " is "; print_Channel b; newline 0)
and print_Val e =
        let val d = !DETAILS in DETAILS := false; print_Exp e; DETAILS := d end
and print_Type ty = (case ty of
	  BoolType => print "bool"
	| NatType => print "nat"
	| IntType => print "int"
	| NamedType a => print a
	| IntervalType(a,b) => (print_Val a; print ".."; print_Val b)
	| UnionType a => print_list ("union "," | "," end") (fn (xl,BotType) => print_list ("",",","") print xl
							     | (xl,y) => (print_list ("",",","") print xl; print " of "; print_Type y)) a
	| RecordType [] => (* for printing type errors *) print "record ... end"
	| RecordType a 	=> print_list ("record ",", "," end") (fn (xl,y) => (print_list ("",",","") print xl; print ":"; print_Type y)) a
	| ArrayType (e,b) => (print "array "; print_Val e; print " of "; print_Type b)
	| QueueType (IntExp (~1,_),b) => (print "queue "; print "?"; print " of "; print_Type b)
	| QueueType (e,b) => (print "queue "; print_Val e; print " of "; print_Type b)
	(* for printing type errors: *)
	| AnyArray => print "array ... of ..."
	| AnyQueue => print "queue ... of ..."
	| BotType  => if !DETAILS then print "0" else print "?"
	| TopType => if !DETAILS then print "1" else print "?")
and print_Channel ch = (case ch of
          [] => print "sync" (* none obsolete *)
	| tyl => print_list (""," # ","") print_Type tyl)
and print_ConstantDecl i (_,n,ty,e) =
        (newline 0; print "const "; print n; print " : "; print_Type ty;
	 print " is "; print_Exp e; print "\n")
and print_ProcessDecl i (_,a,b,d,e,f,p,j) =
	(newline 0; print "process "; print a;
	 case b of [] => () | _ => (newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline (i+4); print_list ("(",", ",")") print_VarParam d);
	 newline (i+4); print "is";
         newline (i+4); print_StatesDecl e;
	 case f of [] => () | _ => print_locVarDecls (i+4) f;
	 case p of SOME (_,s) => (newline (i+4); print "init "; print_Statement (i+7) s) | _ => ();
	 app (print_Transition (i+4)) j;
	 newline 0)
and print_FArgs (vals,ty) =
	(print_list ("",",","") (fn (_,n) => print n) vals; print ": "; print_Type ty)
and print_FunctionDecl i (_,a,b,c,d,e) =
	(newline 0; print "function "; print a;
	 print_list (" (",", ",")") print_FArgs b;
	 print " : "; print_Type c;
	 newline (i+4); print "is";
	 case d of [] => () | _ => print_locVarDecls (i+4) d;
	 newline (i+4);
	 case e of StReturn (_,exp) => (print "return "; print_Exp exp)
		 | _ => (print "begin"; print_Statement (i+6) e; newline (i+4); print "end");
	 newline 0)
and print_ExternDecl i (_,a,b,c,d) =
	(newline 0; print "extern "; print a;
	 print_list (" (",", ",")") print_Type b;
	 print " : "; print_Type c;
	 print " is "; print d;
	 newline 0)
and print_ComponentDecl i (_,a,b,d,e,f,g,p,h) =
	(newline 0; print "component "; print a;
	 case b of [] => () | _ => (newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline (i+4); print_list ("(",", ",")") print_VarParam d);
	 newline (i+4); print "is";
	 case e of [] => () | _ => print_locVarDecls (i+4) e;
	 case f of [] => () | _ => print_locPortDecls (i+4) f;
         case g of [] => () | _ => (newline (i+4); print_list ("priority ",", ","") print_PriorDecl g);
	 case p of SOME (_,s) => (newline (i+4); print "init "; print_Statement (i+7) s) | _ => ();
	 newline (i+4); print_Composition (i+4) h true;
	 newline 0)
and print_locPortDecls i (h::t) =
        let fun printdec (dec as (pals,profile),prb,(left,right)) =
		(print_PortDecl dec; print_TimeConstraint (left,right);
		 case prb of SOME pb => (print " is "; print_Probe pb) | _ => ())
	in newline i; print "port "; printdec h;
	   if null t then () else List.app (fn d => (print ","; newline (i+5); printdec d)) t
	end
and print_Probe pb = (case pb of
          StateProbe (p,s) => (print_Path p; print "state "; print s)
        | ValueProbe (p,e) => (print_Path p; print "value "; print_Exp e)
        | EventProbe (p,r) => (print_Path p; print "event "; print r)
        | EnterProbe r => (print "enter "; print_Probe r)
        | LeaveProbe r => (print "leave "; print_Probe r)
	| NotProbe r => (print "not "; print_Probe r)
	| AndProbe (r1,r2) => (print_Probe r1; print " and "; print_Probe r2)
	| OrProbe (r1,r2) => (print_Probe r1; print " or "; print_Probe r2)
	| ImpProbe (r1,r2) => (print_Probe r1; print " => "; print_Probe r2)
	| ParenProbe (r as ParenProbe _) => print_Probe r
	| ParenProbe r => (print "("; print_Probe r; print ")")
        (* as parsed *)
	| ThatProbe (s,path,e) => (print_Path path; print s; print " "; print_Exp e)
	| JuxtProbe (s,p) => (print s; print " "; print_Probe p))
and print_Path (root,p) =
        (print root; print "/"; case p of [] => () | _ => print_list ("","/","/") (print o Int.toString) p)
and print_lasting (Closed r) = (print " lasting "; print (Int.toString (Real.floor r)))
  | print_lasting _ = print "?"
and print_duration (SOME (Closed r)) = (print " for "; print (Int.toString (Real.floor r)))
  | print_duration (SOME _) = (print " for "; print "?")
  | print_duration _ = ()
and print_within (SOME i) = (print " within "; print_Interval i)
  | print_within _ = ()
and print_Property p = (case p of
       (* general *)
         DeadlockFreeProp => print "deadlockfree"
       | InfinitelyOftenProp p => (print "infinitelyoften "; print_Probe p)
       | MortalProp p => (print "mortal "; print_Probe p)
       | LiveProp p => (print "live "; print_Probe p)
       (* presence *)
       | PresentProp (p,i) => (print "present "; print_Probe p; print_within i)
       | PresentLastingProp (p,d) => (print "present "; print_Probe p; print_lasting d)
       | PresentAfterProp (p1,p2,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print_within i)
       | PresentBeforeProp (p1,p2,i) => (print "present "; print_Probe p1; print " before "; print_Probe p2; print_within i)
       | PresentBetweenProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | PresentUntilProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       (* absence *)
       | AbsentProp (p,i) => (print "absent "; print_Probe p; print_within i)
       | AbsentLastingProp (p,d) => (print "absent "; print_Probe p; print_lasting d)
       | AbsentAfterProp (p1,p2,d) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print_duration d)
       | AbsentBeforeProp (p1,p2,d) => (print "absent "; print_Probe p1; print " before "; print_Probe p2; print_duration d)
       | AbsentBetweenProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AbsentUntilProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       (* response *)
       | LeadstoProp (p1,p2,i) => (print_Probe p1; print "leadsto "; print_Probe p2; print_within i)
       | LeadstoBeforeProp (p1,p2,p3,i) => (print_Probe p1; print "leadsto "; print_Probe p2; print " before "; print_Probe p3; print_within i)
       | LeadstoAfterProp (p1,p2,p3,i) =>  (print_Probe p1; print "leadsto "; print_Probe p2; print " after "; print_Probe p3; print_within i)
       | LeadstoBetweenProp (p1,p2,p3,p4,i) => (print_Probe p1; print "leadsto "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_within i)
       | LeadstoUntilProp (p1,p2,p3,p4,i) => (print_Probe p1; print "leadsto "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_within i)
       (* universality *)
       | AlwaysProp (p,i) => (print "always "; print_Probe p; print_within i)
       | AlwaysLastingProp (p,b) => (print "always "; print_Probe p; print_lasting b)
       | AlwaysAfterProp (p1,p2,b) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print_duration b)
       | AlwaysBeforeProp (p1,p2,b) => (print "always "; print_Probe p1; print " before "; print_Probe p2; print_duration b)
       | AlwaysBetweenProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AlwaysUntilProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       (* precedence *)
       | PrecedesProp (p1,p2,b) => (print_Probe p1; print " precedes "; print_Probe p2; print_duration b)
       | PrecedesBeforeProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " before "; print_Probe p3; print_duration b)
       | PrecedesAfterProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print_duration b)
       | PrecedesBetweenProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_duration b)
       | PrecedesUntilProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_duration b)
       (* composite *)
       | AndProp (p1,p2) => (print_Property p1; print " and "; print_Property p2)
       | OrProp (p1,p2) => (print_Property p1; print " or "; print_Property p2)
       | NotProp p => (print "not "; print_Property p)
       | ImpProp (p1,p2) => (print_Property p1; print " => "; print_Property p2)
       (* ltl props *)
       | BoxProp p => (print "[] "; print_Property p)
       | DiaProp p => (print "<> "; print_Property p)
       | NextProp p => (print "() "; print_Property p)
       | LtlProp p => (print "ltl "; print_Property p)
       | UntilProp (p1,p2) => (print_Property p1; print " until "; print_Property p2)
       | ReleaseProp (p1,p2) => (print_Property p1; print " release "; print_Property p2)
       | ProbeProp p => print_Probe p
       (* parenthesis *)
       | ParenProp p => (print "("; print_Property p; print ")")
       (* variable *)
       | VarProp (line,n) => print n)
and print_PropertyDecl i (_,a,p) =
        (newline 0; print "property "; print a; print " is "; print_Property p)
and print_PortSet pl = (case pl of
	  AllPorts => print "*"
	| SomePorts l => print_list ("",",","") print l)
and print_Composition i c x = (case c of (* x true if starts at beginning of line *)
	  ParComp (_,pl,cl)  => (if x then () else newline i; print "par";
				 case pl of SomePorts [] => () | _ => (print " "; print_PortSet pl; print " in");
				 let val (pl,h)::t = cl
				     val no = ref false
				 in newline (i+3);
				    case pl of SomePorts [] => no:=true | _ => (print_PortSet pl; print " -> ");
				    print_Composition (i+3) h (!no);
				    List.app (fn (pl,c) =>
						(newline i; print "|| ";
						 case pl of SomePorts [] => no:=true | _ => (print_PortSet pl; print " -> ");
						 print_Composition (i+3) c (!no))) t
				 end;
				 newline i; print "end")
	| InstanceComp c => print_InstanceComp i c)
and print_InstanceComp i ((_,id),pl,al) =
	(print id;
	 if null pl then () else print_list (" [",",","]") print pl;
	 if null al then () else print_list (" (",",",")") print_Exp al)
and print_PortDecl (pals,profile) =
	(print_list ("",",","") (print o snd) pals; print ": "; print_Profile profile)
and print_Attrs attrs =
        print_list ("","","") (fn IN => print "in " | OUT => print "out ") attrs
and print_Profile (al,ch) = (case ch of
          [] => print "sync"
	| _  => (print_Attrs al; print_Channel ch))
and print_VarParam (vals,attrl,ty) =
	(print_list ("",",","") (fn (_,RefArg n) => (print "&"; print n) | (_,ValArg n) => print n) vals; print ": ";
	 if null attrl then () else print_list (""," ","") (fn READ => print "read " | WRITE => print "write ") attrl;
	 print_Type ty)
and print_locVarDecls i (h::t) =
        let fun printvar (a,b,c) =
		(print_list ("",",","") (print o snd) a; print ": "; print_Type b;
		 case c of SOME ival => (print " := "; print_Exp ival) | _ => ())
        in newline i; print "var "; printvar h;
	   List.app (fn d => (print ","; newline (i+4); printvar d)) t
	end
and print_StatesDecl (_,a) =
	(print_list ("states ",", ","") print a)
and print_PriorDecl (a,b) =
	(print_list ("","|","") (print o snd) a; print " > "; print_list ("","|","") (print o snd) b)
and print_TimeConstraint (left,right) =
        if case (left,right) of (Closed d,Infinite) => d <= 0.0 | _ => false then () else
	(print " in "; print_Interval (left,right))
and print_Interval (left,right) =
	(case left of Open b => (print "]"; print (Real.toString b))
		    | Closed b => (print "["; print (Real.toString b));
	 print ",";
	 case right of Open b => (print (Real.toString b); print "[")
		    | Closed b => (print (Real.toString b); print "]")
		    | Infinite => print "...[")
and print_Transition i (_,b,c) =
	(newline i; print "from "; print b; print " "; print_Statement (i+3) c)
and print_Statement e st = case st of
	  StNull => (newline e; print "null")
	| StTag (_,s) => if !STRIP then () else (newline e; print "# "; print s)
	| StCond(l,a,b,c,els) =>
		(newline e; print "if "; print_Exp a; print " then"; print_Statement (e+3) b;
		 app (fn (_,a,b) => (newline e; print "elsif "; print_Exp a; print " then"; print_Statement (e+3) b)) c;
		 case els of SOME d => (newline e; print "else";  print_Statement (e+3) d) | _ => ();
		 newline e; print "end")
	| StSelect (a::t) =>
	  let fun print_alternant e (a::t) =
		  (print_Statement (e+3) a;
		   List.app (fn y => (newline e; print "[]"; print_Statement (e+3) y)) t);
	  in newline e; print "select";
	     print_alternant e a;
	     List.app (fn y => (newline e; print "unless"; print_alternant e y)) t;
	     newline e; print "end"
	  end
	| StSequence(_,a,b) => (print_Statement e a; print "; "; print_Statement e b)
	| StSignal (_,a,r) => (newline e; print a; print_pdetails (!r))
	| StInput (_,a,r,b,p) =>
		(newline e; print a; print_pdetails (!r); print "? "; print_list ("",",","") print_Patt b;
		 case p of SOME c => (print " where "; print_Exp c) | _ => ())
 	| StOutput (_,a,r,b) =>
		(newline e; print a; print_pdetails (!r); print "! "; print_list ("",", ","") print_Exp b)
	| StAssign (l,a,b) =>
		(newline e; print_list ("",",","") print_Patt a; print " := "; print_list ("",",","") print_Exp b)
	| StAny (l,a,r,w) => (print_list ("",",","") print_Patt a; print_rdetails (!r); print ":= any";
		case w of SOME d => (print " where "; print_Exp d) | _ => ())
	| StWhile(l,a,b) => (newline e; print "while "; print_Exp a; print " do"; 
			     print_Statement (e+3) b; 
			     newline e; print "end")
	| StCase (_,a,psl) =>
	        let fun printMatches ((_,ref b,a1,s1)::psl) =
			(newline (e+2); print_Patt a1; print " -> "; print_Statement (e+3) s1;
			 List.app (fn (_,ref b,a,s) => (newline e; print "| "; print_Patt a; print " -> "; print_Statement (e+3) s)) psl)
		in newline e; print "case "; print_Exp a; print " of";
		   printMatches (if !TRANSF then List.filter (fn (_,ref b,_,_) => b) psl else psl);
		   newline e; print "end"
		end
	| StForeach (_,v,t,s) => (newline e; print "foreach "; print v; print_details (!t); print " do ";
				  print_Statement (e+3) s;
				  newline e; print "end")
	| StTo (l,a) => (newline e; case a of "" => print "loop" | _ => (print "to "; print a))
        | StOn (l,exp) => (newline e; print "on "; print_Exp exp)
	| StWait (_,l,r) => (newline e; print "wait "; print_Interval (l,r))
	| StReturn (_,exp) => (newline e; print "return "; print_Exp exp)
and print_Patt e =
        let val d = !DETAILS in DETAILS := false; print_Exp e; DETAILS := d end
and is_atomic e = case e of
	  ArrayAccessExp _ => true
	| RecordAccessExp _ => true
	| IdentExp _ => true
	| BoolExp _ => true
	| IntExp _ => true
	| _ => false
and print_Exp e = case e of
	  IdentExp (a,ref (_,ty)) => (print a (* ; print_details ty *))
	| AnyExp 		=> print "any"
	| ArrayAccessExp(a,b,_) => (print_Exp a; print "["; print_Exp b; print "]")
 	| RecordAccessExp(a,b,_) => (print_Exp a; print "."; print b)
	| IntExp (n,r)          => (printi n; print_details (!r))
	| BoolExp b             => print (Bool.toString b)
	| BinopExp(x,r,y,z)	=> (print_Binop x; print_details (!r); print "("; print_Exp y; print ","; print_Exp z; print ")")
	| UnopExp(a,r,b)        => (print_Unop a; print_details (!r); print " "; print_Exp b)
	| ConstrExp(c,e2,r)     => (print c; print " "; print_Exp e2; print_details (!r))
	| JuxtExp(e,[ParenExp e'],r) => print_Exp (JuxtExp(e,[e'],r))
	| JuxtExp(e,[e'],r)     => (print_Exp e; print " "; if is_atomic e' then print_Exp e' else (print "("; print_Exp e'; print ")"))
	| JuxtExp(e,el,_)       => (print_Exp e; print " "; print_list ("(",",",")") print_Exp el)
	| RecordExp (fl,r)      => let val b = !DETAILS
				   in DETAILS := false; print_list ("{",",","}") (fn (f,l) => (print f; print "="; print_Exp l)) fl;
				      DETAILS := b; print_details (!r)
				   end
	| ArrayExp (el,r)       => let val b = !DETAILS
				   in DETAILS := false; print_list ("[",",","]") print_Exp el; DETAILS := b; print_details (!r)
			           end
	| QueueExp (el,r)   	=> (print_list ("{|",",","|}") print_Exp el; print_details (!r))
	| RefExp (v,r)          => (print "&"; print v; print_details (!r))
	| ParenExp a	        => (print "("; print_Exp a; print ")")
	| _ => (* parenthesizes infixes and conditionals if -p; helps check parsing *)
	  (if !TYPECHECK then () else print "(";
	   case e of
          InfixExp(x,r,y,z) => (print_Exp y; print " "; print_Infix x; print_details (!r); print " "; print_Exp z)
	| CondExp (c,e1,e2,r) => (print_Exp c; print " ? "; print_Exp e1; print " : "; print_Exp e2);
	   if !TYPECHECK then () else print ")")
and print_details ty =
	if !DETAILS then (print "::"; print_Type ty) else ()
and print_rdetails tyl =
	if !DETAILS then (print "::"; print_list (""," * ","") print_Type tyl) else ()
and print_pdetails ty =
	if !DETAILS then (print "::"; print_Profile ty) else ()
and print_Infix ope = print (case ope of
	  AND => "and"	| OR  => "or"	| IMP => "=>"   | LE  => "<="	| LT  => "<"	| GT  => ">"	| GE  => ">="
	| EQ  => "="	| NE  => "<>"	| ADD => "+"	| SUB => "-"	| MOD => "%"	| MUL => "*"	| DIV => "/")
and print_Binop ENQUEUE = print "enqueue" | print_Binop APPEND = print "append"
and print_Unop ope = print (case ope of
	  COERCE => "$" | NOT => "not" | FULL => "full" | EMPTY => "empty" | DEQUEUE => "dequeue" | PLUS => "+" | MINUS => "-" | FIRST => "first" | LENGTH => "length")
;

fun printFiacre s0 = (print_Fiacre s0; newline 0; flush ())
   
end




structure ErrSrcPrint = SrcPrint(ErrPrint);
structure OutSrcPrint = SrcPrint(OutPrint);

signature PRINT = sig val print : Fiacre -> unit end;

structure Print : PRINT = struct
fun print st = case !OUTPUT of
	  Fiacre => OutSrcPrint.printFiacre st
	| TTS => ()
	| LOTOS => ()
	| _ => ()
end;

